@@ -96,7 +96,8 @@ type partitionProducer struct {
 	topic  string
 	log    log.Logger
 
-	conn uAtomic.Value
+	conn         uAtomic.Value
+	cnxKeySuffix int32
 
 	options                  *ProducerOptions
 	producerName             string
@@ -179,6 +180,7 @@ func newPartitionProducer(client *client, topic string, options *ProducerOptions
 		client:           client,
 		topic:            topic,
 		log:              logger,
+		cnxKeySuffix:     client.cnxPool.GenerateRoundRobinIndex(),
 		options:          options,
 		producerID:       client.rpcClient.NewProducerID(),
 		dataChan:         make(chan *sendRequest, maxPendingMessages),
@@ -301,7 +303,7 @@ func (p *partitionProducer) grabCnx(assignedBrokerURL string) error {
 		cmdProducer.Metadata = toKeyValues(p.options.Properties)
 	}
 
-	cnx, err := p.client.cnxPool.GetConnection(lr.LogicalAddr, lr.PhysicalAddr)
+	cnx, err := p.client.cnxPool.GetConnection(lr.LogicalAddr, lr.PhysicalAddr, p.cnxKeySuffix)
 	// registering the producer first in case broker sends commands in the middle
 	if err != nil {
 		p.log.Error("Failed to get connection")
@@ -392,7 +394,7 @@ func (p *partitionProducer) grabCnx(assignedBrokerURL string) error {
 			pi.sentAt = time.Now()
 			pi.Unlock()
 			p.pendingQueue.Put(pi)
-			p._getConn().WriteData(pi.buffer)
+			p._getConn().WriteData(pi.ctx, pi.buffer)
 
 			if pi == lastViewItem {
 				break
@@ -471,8 +473,7 @@ func (p *partitionProducer) getOrCreateSchema(schemaInfo *SchemaInfo) (schemaVer
 
 func (p *partitionProducer) reconnectToBroker(connectionClosed *connectionClosed) {
 	var (
-		maxRetry           int
-		delayReconnectTime time.Duration
+		maxRetry int
 	)
 	if p.options.MaxReconnectToBroker == nil {
 		maxRetry = -1
@@ -482,75 +483,59 @@ func (p *partitionProducer) reconnectToBroker(connectionClosed *connectionClosed
 
 	bo := p.backOffPolicyFunc()
 
-	for maxRetry != 0 {
-		select {
-		case <-p.ctx.Done():
-			return
-		default:
+	var assignedBrokerURL string
+	if connectionClosed != nil && connectionClosed.HasURL() {
+		assignedBrokerURL = connectionClosed.assignedBrokerURL
+	}
+
+	opFn := func() (struct{}, error) {
+		if maxRetry == 0 {
+			return struct{}{}, nil
 		}
 
 		if p.getProducerState() != producerReady {
 			// Producer is already closing
 			p.log.Info("producer state not ready, exit reconnect")
-			return
-		}
-
-		var assignedBrokerURL string
-
-		if connectionClosed != nil && connectionClosed.HasURL() {
-			delayReconnectTime = 0
-			assignedBrokerURL = connectionClosed.assignedBrokerURL
-			connectionClosed = nil // Only attempt once
-		} else {
-			delayReconnectTime = bo.Next()
-		}
-
-		p.log.WithFields(log.Fields{
-			"assignedBrokerURL":  assignedBrokerURL,
-			"delayReconnectTime": delayReconnectTime,
-		}).Info("Reconnecting to broker")
-		time.Sleep(delayReconnectTime)
-
-		// double check
-		if p.getProducerState() != producerReady {
-			// Producer is already closing
-			p.log.Info("producer state not ready, exit reconnect")
-			return
+			return struct{}{}, nil
 		}
 
 		atomic.AddUint64(&p.epoch, 1)
 		err := p.grabCnx(assignedBrokerURL)
+		if assignedBrokerURL != "" {
+			// Only attempt once
+			assignedBrokerURL = ""
+		}
 		if err == nil {
 			// Successfully reconnected
 			p.log.WithField("cnx", p._getConn().ID()).Info("Reconnected producer to broker")
 			bo.Reset()
-			return
+			return struct{}{}, nil
 		}
 		p.log.WithError(err).Error("Failed to create producer at reconnect")
 		errMsg := err.Error()
 		if strings.Contains(errMsg, errMsgTopicNotFound) {
 			// when topic is deleted, we should give up reconnection.
 			p.log.Warn("Topic not found, stop reconnecting, close the producer")
-			p.doClose(joinErrors(ErrTopicNotfound, err))
-			break
+			p.doClose(errors.Join(ErrTopicNotfound, err))
+			return struct{}{}, nil
 		}
 
 		if strings.Contains(errMsg, errMsgTopicTerminated) {
 			p.log.Warn("Topic was terminated, failing pending messages, stop reconnecting, close the producer")
-			p.doClose(joinErrors(ErrTopicTerminated, err))
-			break
+			p.doClose(errors.Join(ErrTopicTerminated, err))
+			return struct{}{}, nil
 		}
 
 		if strings.Contains(errMsg, errMsgProducerBlockedQuotaExceededException) {
 			p.log.Warn("Producer was blocked by quota exceed exception, failing pending messages, stop reconnecting")
-			p.failPendingMessages(joinErrors(ErrProducerBlockedQuotaExceeded, err))
-			break
+			p.failPendingMessages(errors.Join(ErrProducerBlockedQuotaExceeded, err))
+			return struct{}{}, nil
 		}
 
 		if strings.Contains(errMsg, errMsgProducerFenced) {
 			p.log.Warn("Producer was fenced, failing pending messages, stop reconnecting")
-			p.doClose(joinErrors(ErrProducerFenced, err))
-			break
+			p.doClose(errors.Join(ErrProducerFenced, err))
+			return struct{}{}, nil
 		}
 
 		if maxRetry > 0 {
@@ -560,7 +545,17 @@ func (p *partitionProducer) reconnectToBroker(connectionClosed *connectionClosed
 		if maxRetry == 0 || bo.IsMaxBackoffReached() {
 			p.metrics.ProducersReconnectMaxRetry.Inc()
 		}
+
+		return struct{}{}, err
 	}
+	_, _ = internal.Retry(p.ctx, opFn, func(_ error) time.Duration {
+		delayReconnectTime := bo.Next()
+		p.log.WithFields(log.Fields{
+			"assignedBrokerURL":  assignedBrokerURL,
+			"delayReconnectTime": delayReconnectTime,
+		}).Info("Reconnecting to broker")
+		return delayReconnectTime
+	})
 }
 
 func (p *partitionProducer) runEventsLoop() {
@@ -842,6 +837,8 @@ func (p *partitionProducer) internalSingleSend(
 
 type pendingItem struct {
 	sync.Mutex
+	ctx           context.Context
+	cancel        context.CancelFunc
 	buffer        internal.Buffer
 	sequenceID    uint64
 	createdAt     time.Time
@@ -900,14 +897,17 @@ func (p *partitionProducer) writeData(buffer internal.Buffer, sequenceID uint64,
 		return
 	default:
 		now := time.Now()
+		ctx, cancel := context.WithCancel(context.Background())
 		p.pendingQueue.Put(&pendingItem{
+			ctx:          ctx,
+			cancel:       cancel,
 			createdAt:    now,
 			sentAt:       now,
 			buffer:       buffer,
 			sequenceID:   sequenceID,
 			sendRequests: callbacks,
 		})
-		p._getConn().WriteData(buffer)
+		p._getConn().WriteData(ctx, buffer)
 	}
 }
 
@@ -1093,8 +1093,8 @@ func (p *partitionProducer) Send(ctx context.Context, msg *ProducerMessage) (Mes
 	isDone := uAtomic.NewBool(false)
 	doneCh := make(chan struct{})
 
-	p.internalSendAsync(ctx, msg, func(ID MessageID, message *ProducerMessage, e error) {
-		if isDone.CAS(false, true) {
+	p.internalSendAsync(ctx, msg, func(ID MessageID, _ *ProducerMessage, e error) {
+		if isDone.CompareAndSwap(false, true) {
 			err = e
 			msgID = ID
 			close(doneCh)
@@ -1118,18 +1118,18 @@ func (p *partitionProducer) SendAsync(ctx context.Context, msg *ProducerMessage,
 
 func (p *partitionProducer) validateMsg(msg *ProducerMessage) error {
 	if msg == nil {
-		return joinErrors(ErrInvalidMessage, fmt.Errorf("message is nil"))
+		return errors.Join(ErrInvalidMessage, fmt.Errorf("message is nil"))
 	}
 
 	if msg.Value != nil && msg.Payload != nil {
-		return joinErrors(ErrInvalidMessage, fmt.Errorf("can not set Value and Payload both"))
+		return errors.Join(ErrInvalidMessage, fmt.Errorf("can not set Value and Payload both"))
 	}
 
 	if p.options.DisableMultiSchema {
 		if msg.Schema != nil && p.options.Schema != nil &&
 			msg.Schema.GetSchemaInfo().hash() != p.options.Schema.GetSchemaInfo().hash() {
 			p.log.Errorf("The producer %s of the topic %s is disabled the `MultiSchema`", p.producerName, p.topic)
-			return joinErrors(ErrSchema, fmt.Errorf("msg schema can not match with producer schema"))
+			return errors.Join(ErrSchema, fmt.Errorf("msg schema can not match with producer schema"))
 		}
 	}
 
@@ -1142,19 +1142,19 @@ func (p *partitionProducer) prepareTransaction(sr *sendRequest) error {
 	}
 
 	txn := (sr.msg.Transaction).(*transaction)
-	if txn.state != TxnOpen {
-		p.log.WithField("state", txn.state).Error("Failed to send message" +
+	if txn.state.Load() != int32(TxnOpen) {
+		p.log.WithField("state", txn.state.Load()).Error("Failed to send message" +
 			" by a non-open transaction.")
-		return joinErrors(ErrTransaction,
+		return errors.Join(ErrTransaction,
 			fmt.Errorf("failed to send message by a non-open transaction"))
 	}
 
 	if err := txn.registerProducerTopic(p.topic); err != nil {
-		return joinErrors(ErrTransaction, err)
+		return errors.Join(ErrTransaction, err)
 	}
 
 	if err := txn.registerSendOrAckOp(); err != nil {
-		return joinErrors(ErrTransaction, err)
+		return errors.Join(ErrTransaction, err)
 	}
 
 	sr.transaction = txn
@@ -1180,7 +1180,7 @@ func (p *partitionProducer) updateSchema(sr *sendRequest) error {
 	if schemaVersion == nil {
 		schemaVersion, err = p.getOrCreateSchema(schema.GetSchemaInfo())
 		if err != nil {
-			return joinErrors(ErrSchema, fmt.Errorf("get schema version fail, err: %w", err))
+			return errors.Join(ErrSchema, fmt.Errorf("get schema version fail, err: %w", err))
 		}
 		p.schemaCache.Put(schema.GetSchemaInfo(), schemaVersion)
 	}
@@ -1197,7 +1197,7 @@ func (p *partitionProducer) updateUncompressedPayload(sr *sendRequest) error {
 	if sr.msg.Value != nil {
 		if sr.schema == nil {
 			p.log.Errorf("Schema encode message failed %s", sr.msg.Value)
-			return joinErrors(ErrSchema, fmt.Errorf("set schema value without setting schema"))
+			return errors.Join(ErrSchema, fmt.Errorf("set schema value without setting schema"))
 		}
 
 		// payload and schema are mutually exclusive
@@ -1205,7 +1205,7 @@ func (p *partitionProducer) updateUncompressedPayload(sr *sendRequest) error {
 		schemaPayload, err := sr.schema.Encode(sr.msg.Value)
 		if err != nil {
 			p.log.WithError(err).Errorf("Schema encode message failed %s", sr.msg.Value)
-			return joinErrors(ErrSchema, err)
+			return errors.Join(ErrSchema, err)
 		}
 
 		sr.uncompressedPayload = schemaPayload
@@ -1375,59 +1375,59 @@ func (p *partitionProducer) ReceivedSendReceipt(response *pb.CommandSendReceipt)
 		p.log.Warnf("Received ack for %v on sequenceId %v - expected: %v, local > remote, ignore it",
 			response.GetMessageId(), response.GetSequenceId(), pi.sequenceID)
 		return
-	} else {
-		// The ack was indeed for the expected item in the queue, we can remove it and trigger the callback
-		p.pendingQueue.Poll()
+	}
+	// The ack was indeed for the expected item in the queue, we can remove it and trigger the callback
+	p.pendingQueue.Poll()
 
-		now := time.Now().UnixNano()
+	now := time.Now().UnixNano()
 
-		// lock the pending item while sending the requests
-		pi.Lock()
-		defer pi.Unlock()
-		p.metrics.PublishRPCLatency.Observe(float64(now-pi.sentAt.UnixNano()) / 1.0e9)
-		batchSize := int32(len(pi.sendRequests))
-		for idx, i := range pi.sendRequests {
-			sr := i.(*sendRequest)
-			atomic.StoreInt64(&p.lastSequenceID, int64(pi.sequenceID))
+	// lock the pending item while sending the requests
+	pi.Lock()
+	defer pi.Unlock()
+	p.metrics.PublishRPCLatency.Observe(float64(now-pi.sentAt.UnixNano()) / 1.0e9)
+	batchSize := int32(len(pi.sendRequests))
+	for idx, i := range pi.sendRequests {
+		sr := i.(*sendRequest)
+		atomic.StoreInt64(&p.lastSequenceID, int64(pi.sequenceID))
 
-			msgID := newMessageID(
-				int64(response.MessageId.GetLedgerId()),
-				int64(response.MessageId.GetEntryId()),
-				int32(idx),
-				p.partitionIdx,
-				batchSize,
-			)
+		msgID := newMessageID(
+			int64(response.MessageId.GetLedgerId()),
+			int64(response.MessageId.GetEntryId()),
+			int32(idx),
+			p.partitionIdx,
+			batchSize,
+		)
 
-			if sr.totalChunks > 1 {
-				if sr.chunkID == 0 {
-					sr.chunkRecorder.setFirstChunkID(
-						&messageID{
-							int64(response.MessageId.GetLedgerId()),
-							int64(response.MessageId.GetEntryId()),
-							-1,
-							p.partitionIdx,
-							0,
-						})
-				} else if sr.chunkID == sr.totalChunks-1 {
-					sr.chunkRecorder.setLastChunkID(
-						&messageID{
-							int64(response.MessageId.GetLedgerId()),
-							int64(response.MessageId.GetEntryId()),
-							-1,
-							p.partitionIdx,
-							0,
-						})
-					// use chunkMsgID to set msgID
-					msgID = &sr.chunkRecorder.chunkedMsgID
-				}
+		if sr.totalChunks > 1 {
+			switch sr.chunkID {
+			case 0:
+				sr.chunkRecorder.setFirstChunkID(
+					&messageID{
+						int64(response.MessageId.GetLedgerId()),
+						int64(response.MessageId.GetEntryId()),
+						-1,
+						p.partitionIdx,
+						0,
+					})
+			case sr.totalChunks - 1:
+				sr.chunkRecorder.setLastChunkID(
+					&messageID{
+						int64(response.MessageId.GetLedgerId()),
+						int64(response.MessageId.GetEntryId()),
+						-1,
+						p.partitionIdx,
+						0,
+					})
+				// use chunkMsgID to set msgID
+				msgID = &sr.chunkRecorder.chunkedMsgID
 			}
-
-			sr.done(msgID, nil)
 		}
 
-		// Mark this pending item as done
-		pi.done(nil)
+		sr.done(msgID, nil)
 	}
+
+	// Mark this pending item as done
+	pi.done(nil)
 }
 
 func (p *partitionProducer) internalClose(req *closeProducer) {
@@ -1552,7 +1552,7 @@ func (p *partitionProducer) setProducerState(state producerState) {
 // set a new producerState and return the last state
 // returns bool if the new state has been set or not
 func (p *partitionProducer) casProducerState(oldState, newState producerState) bool {
-	return p.state.CAS(int32(oldState), int32(newState))
+	return p.state.CompareAndSwap(int32(oldState), int32(newState))
 }
 
 func (p *partitionProducer) Close() {
@@ -1584,14 +1584,14 @@ type sendRequest struct {
 	uuid             string
 	chunkRecorder    *chunkRecorder
 
-	/// resource management
+	// resource management
 
 	memLimit          internal.MemoryLimitController
 	reservedMem       int64
 	semaphore         internal.Semaphore
 	reservedSemaphore int
 
-	/// convey settable state
+	// convey settable state
 
 	sendAsBatch         bool
 	transaction         *transaction
@@ -1664,7 +1664,7 @@ func (sr *sendRequest) done(msgID MessageID, err error) {
 }
 
 func (p *partitionProducer) blockIfQueueFull() bool {
-	//DisableBlockIfQueueFull == false means enable block
+	// DisableBlockIfQueueFull == false means enable block
 	return !p.options.DisableBlockIfQueueFull
 }
 
@@ -1741,11 +1741,17 @@ func (i *pendingItem) done(err error) {
 	if i.isDone {
 		return
 	}
+
 	i.isDone = true
-	buffersPool.Put(i.buffer)
+	// return the buffer to the pool after all callbacks have been called.
+	defer buffersPool.Put(i.buffer)
 	if i.flushCallback != nil {
 		i.flushCallback(err)
 	}
+
+	if i.cancel != nil {
+		i.cancel()
+	}
 }
 
 // _setConn sets the internal connection field of this partition producer atomically.
